#include "stats.h"

#include "horde_log.h"

#include "net/poolman.h"
#include "ippc/ippc_server.h"
#include "ippc/syscall.h"

#include <nmstl/net>
#include <nmstl/ioevent>
#include <vector>

using namespace nmstl;
using namespace std;

HORDE_NAMESPACE_BEGIN;

static const int STAT_TIMER_PERIOD = 1005;

extern io_event_loop* core_ioloop; 
vector<ippc_handler*> listeners;
callback_timer* timer = NULL;

void stat_timer_fire() {
  if (!listeners.empty()) {
    // build the statistics report
    ippcmsg_statistics statmsg;
    bzero(&statmsg, sizeof(statmsg));

    int n = 0;
    int wrtt_sum = 0;
    int wloss_sum = 0;

    vector< ptr<channel_statistics> > stats = get_poolman()->get_statistics();
    for (vector< ptr<channel_statistics> >::const_iterator iter = stats.begin(); iter != stats.end(); iter++) {
      ptr<channel_statistics> stat = *iter;
      string cname = stat->get_channel_name();
      int loss = stat->get_loss_rate();
      int rtt = stat->get_average_rtt_msec();
      int srate = stat->get_average_send_rate();
      int rrate = stat->get_average_recv_rate();
      int asrate = stat->get_allowed_send_rate();
      
      statmsg.total_send_rate += srate;
      statmsg.total_allowed_send_rate += asrate;
      statmsg.total_recv_rate += rrate;
      wrtt_sum += rtt * asrate;
      wloss_sum += loss * asrate;

      strncpy(statmsg.channel_stats[n].channel_id, cname.c_str(), CHANNEL_ID_MAX_LENGTH);
      statmsg.channel_stats[n].send_rate = srate;
      statmsg.channel_stats[n].allowed_send_rate = asrate;
      statmsg.channel_stats[n].recv_rate = rrate;
      statmsg.channel_stats[n].loss_rate = loss;
      statmsg.channel_stats[n].ave_rtt = rtt;

      n++;
      if (n > STAT_MAX_CHANNELS) break;
    }
    
    if (statmsg.total_allowed_send_rate) {
      statmsg.ave_rtt = wrtt_sum / statmsg.total_allowed_send_rate;
      statmsg.loss_rate = wloss_sum / statmsg.total_allowed_send_rate;
    } else {
      statmsg.ave_rtt = 0;
      statmsg.loss_rate = 0;
    }

    statmsg.nchannels = n;

    // deliver the message to every listener
    int k = 0;
    for (vector<ippc_handler*>::const_iterator iter = listeners.begin(); iter != listeners.end(); iter++) {
      (*iter)->send_message(IPPC_STATISTICS, &statmsg, sizeof(struct ippcmsg_statistics));
      k++;
    }
    LOG_EVENT_DEBUG("statistics") << "stat_timer_fire: delivered statistics to " << k << " listeners.";

    // reschedule the timer
    timer = (new callback_timer(*(core_ioloop), stat_timer_fire));
    timer->arm(ntime::now_plus_msecs(STAT_TIMER_PERIOD));
  } else {    
    // nobody cares, so don't collect stats...
    timer = NULL;

    LOG_EVENT("statistics") << "stat_timer_fire: no listeners, disabling statistics reporting.";
  }
}

vector<ippc_handler*>::iterator listeners_find(ippc_handler* handler) {
  vector<ippc_handler*>::iterator iter = listeners.begin();
  for (; iter != listeners.end(); iter++) {
    if (*iter == handler) break;
  }
  return iter;
}

void add_statistics_listener(ippc_handler* handler) {
  if (listeners_find(handler) == listeners.end()) {
    listeners.push_back(handler);

    if (timer == NULL) stat_timer_fire();
  }
}

void remove_statistics_listener(ippc_handler* handler) {
  vector<ippc_handler*>::iterator iter = listeners_find(handler);
  if (iter != listeners.end()) {
    listeners.erase(iter);
  }
}



HORDE_NAMESPACE_END;
