// -*- mode: C++; -*-
/** @file event_handlers.h
 *
 * All the cross-module event handlers and callbacks. Declared in one
 * place to help keep track of cross-module dependencies.
 */ 
#ifndef __HORDE_EVENTHANDLERS_H__
#define __HORDE_EVENTHANDLERS_H__

#include "horde_namespace.h"

#include <nmstl/ptr>
#include <nmstl/net>
using namespace nmstl;

HORDE_NAMESPACE_BEGIN;
class tx_slot;
class horde_net_packet;
class ippc_handler;

// net -> core
void notify_core_new_tx_slot(ptr<tx_slot> slot);
void notify_core_ack_received(int packet_uid, int rtt);
void notify_core_loss_detected(int packet_uid);
void notify_core_packet_received(horde_net_packet packet, inet_address from_address, int channel_id);
void notify_core_iface_up(inet_address address);
void notify_core_iface_down(inet_address address);

// core -> session
void notify_session_connect_request(string session_id, inet_address peer);
void notify_session_connect_ack(string session_id, inet_address peer);
void notify_session_close_request(string session_id);

void notify_stream_connect_request(string session_id, int local_uid, int remote_uid);
void notify_stream_connect_ack(string stream_id, int local_uid, int remote_uid);
void notify_stream_close_request(string stream_id, int local_uid, int remote_uid);
void notify_stream_data_ack(int stream_id, int adu_id, int frag_start, int frag_len, int rtt);
void notify_stream_data_loss(int stream_id, int adu_id, int frag_start, int frag_len);
void notify_stream_data_recv(int stream_id, int adu_id, int frag_start, int frag_len, int tot_len, const char* data);

// session -> core
void notify_core_session_connected();

// session -> syscall
void notify_syscall_session_connected(ippc_handler* ioh);
void notify_syscall_session_closed(ippc_handler* ioh, bool remote=false);

void notify_syscall_stream_connected(ippc_handler* ioh, string stream_id, int uid);
void notify_syscall_stream_closed(ippc_handler* ioh, string stream_id);
void notify_syscall_stream_choke(ippc_handler* ioh, int stream_uid, bool choked);
void notify_syscall_stream_flushed(ippc_handler* ioh, int stream_uid); 
void notify_syscall_stream_data_ack(ippc_handler* ioh, int stream_uid, int adu_id, int frag_start, int frag_length, int rtt_ms);
void notify_syscall_stream_data_loss(ippc_handler* ioh, int stream_uid, int adu_id, int frag_start, int frag_length);
void notify_syscall_stream_data_recv(ippc_handler* ioh, int stream_uid, int adu_id, int frag_start, int total_length, const char* data, int data_length);

enum session_exception_code_t { EX_SESSION_BAD_ID, EX_SESSION_ILLEGAL_OPERATION, 
				EX_SESSION_INSUFFICIENT_PERMISSION };
void notify_syscall_session_exception(ippc_handler* ioh, session_exception_code_t code, string ex_msg);

enum stream_exception_code_t { EX_STREAM_BAD_ID, EX_STREAM_ILLEGAL_OPERATION, 
			       EX_STREAM_INSUFFICIENT_PERMISSION, EX_STREAM_INTERNAL_ERROR };
void notify_syscall_stream_exception(ippc_handler* ioh, string stream_id, stream_exception_code_t code, string ex_msg);

HORDE_NAMESPACE_END;

#endif
