// -*- mode: C++; -*-

/** @file txslot.h
 *
 * The abstractions used between the imux and the network channel
 * managers: transmission slots.
 */

#ifndef __HORDE_TXSLOT_H__
#define __HORDE_TXSLOT_H__

#include "horde_namespace.h"

HORDE_NAMESPACE_BEGIN;

class network_channel_manager;

/**
 * A transmission slot.
 */
class tx_slot {
protected:
  network_channel_manager* parent_manager; ///< channel manager that generated this slot
  int loss_prob;                   ///< estimated loss probability
  int expect_rtt;                  ///< expected round-trip-time
  size_t max_size;                 ///< Maximum number of bytes this slot can transmit
  bool valid;                      ///< can this slot be used to make a transmission? -- only mutable field
 
public:
  tx_slot(network_channel_manager* creator, int rtt, int loss, size_t max_size) :
    parent_manager(creator), loss_prob(loss), expect_rtt(rtt), max_size(max_size), valid(true)
  {  }  
    
  /** Returns a pointer to the channel manager that generated this slot */
  network_channel_manager* get_channel() { return parent_manager; }
    
  /** Returns the loss probability p as a percentage (0 <= p <= 100) */
  int get_loss_probability() { return loss_prob; }

  /** Returns the expected round-trip-time for this slot, in milliseconds. */
  int get_expected_rtt() { return expect_rtt; }
    
  /** Returns the max number of data bytes this slot can hold. */
  size_t get_max_size() { return max_size; }

  /** Returns true iff this slot can be used to make a transmission. */
  bool is_valid() { return valid; }

  /** Makes this slot invalid. Only mutator. */
  void invalidate() { valid = false; }
};

HORDE_NAMESPACE_END;

#endif
