#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#include <signal.h>

#include "ippc.h"
#include "syscall.h"

static volatile int quit = 0;

static void _sigint(int s) {
  printf("signal: %d\n", s);
  quit = 1;
}

void sleep(int timeout_ms) {
  timeval timeout;
  gettimeofday(&timeout, 0);
  timeout.tv_sec = (long)(timeout_ms) / 1000;;
  timeout.tv_usec = ((long)(timeout_ms) % 1000) * 1000;

  select(0, NULL, NULL, NULL, &timeout);
}

int main(void) {
  // setup signal handlers
  signal(SIGINT, _sigint);
  signal(SIGTERM, _sigint);
  signal(SIGHUP, _sigint);
  
  siginterrupt(SIGINT, 1);
  siginterrupt(SIGTERM, 1);
  siginterrupt(SIGHUP, 1);
  
  // test code
  ippc_channel ioc = ippc_channel_open("/var/horde/sessions/test/ippc");
				       //"/tmp/test.ippc");
  if (ioc == INVALID_CHANNEL) {
    perror("ippc_channel_open");    
  } else {
    ippcmsg_pingpong ping;
    ping.seq = 0;
    int rtype;
    size_t rlength;

    while (quit == 0) {
      sleep(1000);

      printf("ping: %d\n", ping.seq);
      char *msg = ippc_call(ioc, IPPC_PING, &ping, sizeof(ping), IPPC_PONG, &rtype, &rlength);
      
      if (msg == NULL) {
	perror("ippc_call");
	quit = 1;
      } else {
	if (rtype == IPPC_PONG) {
	  ippcmsg_pingpong* pong = (ippcmsg_pingpong*)msg;
	  long* x = (long*)msg;
	  printf("pong: %d [%lx, %lx, %lx %lx], %d\n", pong->seq, *x, *(x+1), *(x+2), *(x+3), rlength);  
	} else {
	  // exception
	  ippcmsg_exception *ex = (ippcmsg_exception*)msg;
	  printf("exception %d: '%s'\n", ex->code, ex->msg);
	}
      }
      ping.seq++;
    }
  }
}
