import horde_api
import sys
from time import time

SESSION_NAME = "elastic"
STREAM_NAME = "data"

hstream = 0
npings = 8

def session_cb(x):
    if (x == horde_api.EVENT_SESSION_CLOSED):
        print "session closed (%d). exiting." % x
        horde_api.session_close(SESSION_NAME)
        sys.exit(0)
    else:
        print "unknown session event:", x

def event_cb(event, stream_handle, a, b, c, d):
    if (event == horde_api.EVENT_ADU_FRAG_ACK):
        pass
        print "\tACK: stream=%d, adu=%d, offset=%d, fraglen=%d, rtt=%d" % (stream_handle, a, b, c, d)
    elif (event == horde_api.EVENT_ADU_FRAG_LOSS):
        print "\tLOSS: stream=%d, adu=%d, offset=%d, fraglen=%d" % (stream_handle, a, b, c)
    elif (event == horde_api.EVENT_STREAM_FLUSHED):
        print "\tFLUSH: stream flushed:", stream_handle
    elif (event == horde_api.EVENT_STREAM_CLOSED):
        print "\tCLOSED: stream closed:", stream_handle
    else:        
        print "\tEVENT: Unknown stream event [%d] on stream %d, a=%d, b=%d, c=%d, d=%d" % (event, stream_handle, a, b, c, d)

def send_data_packet(hstream, seq, datalen):
    data = "0" * datalen
    horde_api.stream_send(hstream, seq, data)

def recv_cb(stream_handle, adu_id, totlen, offset, data):    
    print "RECV: data='%s...', adu=%d, totlen=%d, fraglen=%d, offset=%d" % (data[0:8], adu_id, totlen, len(data), offset)

def rate_to_kbps(rate):
    kbps = (rate * 8.0) / 1024;
    return "%.1f kbps" % kbps

def stat_cb(total_send, total_recv, total_allowed_send, nchannels):
    print "\tstatistics: send=%d (%s), recv=%d (%s), allowed=%d (%s), nchannels=%d"\
          % (total_send, rate_to_kbps(total_send), total_recv, rate_to_kbps(total_recv),
             total_allowed_send, rate_to_kbps(total_allowed_send), nchannels)

def horde_event_loop():
    while (horde_api.event_dispatch() > 0):
        pass

def do_client(server):
    global hstream
    
    if (horde_api.session_connect(server, SESSION_NAME)):
        print "connected to server '%s'." % server
        h = horde_api.stream_connect(STREAM_NAME)
        if (h > 0):
            hstream = h
            print "stream '%s' connected, uid:%d" % (STREAM_NAME, h)
            
            # register callbacks
            horde_api.session_set_callback(session_cb)
            horde_api.stream_set_event_callback(h, event_cb)
            horde_api.stream_set_recv_callback(h, recv_cb)

            horde_api.statistics_set_callback(stat_cb)
            horde_api.statistics_enable(1)

            # set stream options
            horde_api.stream_set_options(hstream,
                                         horde_api.STREAM_OPT_IO_PRESSURE
                                         | horde_api.STREAM_OPT_EVENT_LOSS
                                         | horde_api.STREAM_OPT_EVENT_RECV)
#                                         | horde_api.STREAM_OPT_EVENT_ACK)

            # just send packets and make sure the I/O pressure holds us back
            adu_id = 0
            k = 1
            while (k >= 0):
                t = time()
                send_data_packet(hstream, adu_id, 512)
                adu_id += 1
                a = time() - t

                t = time()
                k = horde_api.event_dispatch(0)
                b = time() - t

                #print "a=%.3f, b=%.3f, k=%d" % (a,b,k)
    
def do_server():
    global hstream
    
    if (horde_api.session_accept(SESSION_NAME)):
        print "client connected."
        h = horde_api.stream_accept(STREAM_NAME)
        if (h > 0):
            hstream = h
            print "stream '%s' connected, uid:%d" % (STREAM_NAME, h)
            
            # register callbacks
            horde_api.session_set_callback(session_cb)
            horde_api.stream_set_recv_callback(h, recv_cb)

            horde_api.statistics_set_callback(stat_cb)
            horde_api.statistics_enable(1)

            # client will send the data

            # enter event loop
            horde_event_loop()
            
if len(sys.argv) == 1:
    # server
    do_server()
elif len(sys.argv) == 2:
    # client
    do_client(sys.argv[1])
else:
    print "usage:", sys.argv[0], "<server address>"
