#
# Code to test out the advanced features of the Horde API. Particularly
# focuses on testing out the grunt scheduler.
#
import horde_api
import sys
from time import sleep

#
# globals
#
SESSION_NAME = "grunt"
STREAM_NAMES = ["a","b","c","d"]

ADU_SIZE = 256
ADU_DATA = "HORDE" * (ADU_SIZE/4)
#ADU_DATA = ".".join([str(x) for x in range(64)])
NUM_ADU_PER_TEST = 300
MTX_TEST_PERIOD = 0#0.050    # 50 msec delay in the MTX test

stream_send_counts = {}
stream_ack_counts = {}
stream_rtt_sums = {}
stream_loss_counts = {}

stream_acks = {}

#
# utils
# 
def stream_close_multiple(name_list, h_list):
    n = 0
    for i in range(len(name_list)):
        if h_list[i] > 0:
            horde_api.stream_close(name_list[i], h_list[i]);
            n += 1
    return n

def rate_to_kbps(rate):
    kbps = (rate * 8.0) / 1024;
    return "%.1f kbps" % kbps

def dispatch_events(timeout=-1):
    while True:
        k = horde_api.event_dispatch(0)
        if k == 0: break
        elif k < 0:
            print "ERROR: event_dispatch returned < 0. Exiting."
            sys.exit(0)

#
# client
#
def do_client(server):
    if (horde_api.session_connect(server, SESSION_NAME)):
        print "connected to server '%s'." % server
        hlist = [None]*4
        n = horde_api.stream_connect_multiple(4, STREAM_NAMES, hlist)
        print "connected:", n, "streams"
        if (n == 4):
            # register callbacks
            for h in hlist: horde_api.stream_set_event_callback(h, stream_event_cb)
            horde_api.statistics_set_callback(stat_cb)
            horde_api.statistics_enable(1)

            # run tests
            test_baseline(hlist)
            test_priority(hlist)
            test_cgroup(hlist)
            test_loss_thresh(hlist)
            test_mtx(hlist)
            test_all(hlist)             
        else:
            print "critical: wrong number of streams."

        stream_close_multiple(STREAM_NAMES, hlist)            
        horde_api.session_close(SESSION_NAME)
        print "session closed."


def stream_event_cb(event, hstream, adu_id, frag_offset, frag_len, rtt):
    if event == horde_api.EVENT_ADU_FRAG_ACK:
        global stream_ack_counts, stream_rtt_sums, stream_acks
        stream_ack_counts[hstream] += 1
        stream_rtt_sums[hstream] += rtt
        stream_acks[hstream][adu_id] = True
    elif event == horde_api.EVENT_ADU_FRAG_LOSS:
        global stream_loss_counts
        stream_loss_counts[hstream] += 1
    elif event == horde_api.EVENT_STREAM_CLOSED:
        pass

def stat_cb(total_send, total_recv, total_allowed_send, nchannels):
    print "\rstatistics: send=%d (%s), recv=%d (%s), allowed=%d (%s), nchannels=%d      "\
          % (total_send, rate_to_kbps(total_send), total_recv, rate_to_kbps(total_recv),
             total_allowed_send, rate_to_kbps(total_allowed_send), nchannels),
    sys.stdout.flush()

def _test_init(hlist, testname):
    global stream_send_counts, stream_ack_counts, stream_rtt_sums, stream_loss_counts, stream_acks

    stream_send_counts = {}
    stream_ack_counts = {}
    stream_rtt_sums = {}
    stream_loss_counts = {}
    stream_acks = {}

    for h in hlist:
        stream_send_counts[h] = 0
        stream_ack_counts[h] = 0
        stream_loss_counts[h] = 0
        stream_rtt_sums[h] = 0
        stream_acks[h] = {}

    print "TESTING:", testname

def _test_done(hlist):
    print
    print "RESULTS:"
    print "\tstream\thstream\tsent\tack'd\tno-ack\tlost\tloss %\tave rtt (ms)"

    for i in range(len(hlist)):
        #lost = stream_loss_counts[hlist[i]]
        lost = stream_send_counts[hlist[i]] - len(stream_acks[hlist[i]]) 
        loss_rate = 100.0 * lost / stream_send_counts[hlist[i]]
        if stream_ack_counts[hlist[i]]:
            ave_rtt = stream_rtt_sums[hlist[i]] / stream_ack_counts[hlist[i]]
        else:
            ave_rtt = 0
        
        print "\t%s\t%d\t%d\t%d\t%d\t%d\t%.1f\t%d"\
              % (STREAM_NAMES[i], hlist[i], stream_send_counts[hlist[i]],
                 stream_ack_counts[hlist[i]], lost, stream_loss_counts[hlist[i]],
                 loss_rate, ave_rtt)
    print

def data_send(hstream, adu_id, pri, cgroup, loss_thresh, mtx):
    stream_send_counts[hstream] += 1

    horde_api.stream_send(hstream, adu_id, ADU_DATA, priority=pri,
                          fragment_allowed=0, correlation_group=cgroup,
                          multiple_tx=mtx, relative_loss_threshold=loss_thresh)

def test_baseline(hlist):
    """
    no objectives
    """
    _test_init(hlist,"No objectives")
    
    for adu in range(NUM_ADU_PER_TEST):
        for h in hlist:
            data_send(h, adu, 0, 0, 0, 0)

        dispatch_events(0)

    # drain stream queues
    sleep(10)
    dispatch_events(500)    
    
    _test_done(hlist)

def test_priority(hlist):
    """
    priority = stream_index
    """
    _test_init(hlist,"Priority")
    
    for adu in range(NUM_ADU_PER_TEST):
        pri = 0
        for h in hlist:
            data_send(h, adu, pri, 0, 0, 0)
            pri += 1

        dispatch_events(0)

    sleep(10) # drain stream queues
    dispatch_events(500)
    
    _test_done(hlist)

def test_cgroup(hlist):
    """
    cgroup = ( adu_num + stream_index ) mod 3 + 1

    +1 to avoid cgroup of 0
    """
    _test_init(hlist,"Correlation groups")

    for adu in range(NUM_ADU_PER_TEST):
        i = 0
        for h in hlist:
            cgroup = (adu + i) % 3 + 1
            data_send(h, adu, 0, cgroup, 0, 0)
            i += 1

        dispatch_events(0)

    sleep(10) # drain stream queues
    dispatch_events(500)
    
    _test_done(hlist)

def test_loss_thresh(hlist):
    """
    relative loss threshold of 2% on 'a'
    relative loss threshold of 5% on 'c'
    """
    _test_init(hlist,"Relative loss threshold")
    
    for adu in range(NUM_ADU_PER_TEST):
        i = 0
        for h in hlist:
            if (i == 0): thresh = 2
            elif (i == 2): thresh = 5
            else: thresh = 0
            data_send(h, adu, 0, 0, thresh, 0)
            i += 1

        dispatch_events(0)        

    sleep(10) # drain stream queues
    dispatch_events(500)    
    
    _test_done(hlist)

def test_mtx(hlist):
    """
    odd stream_index's have MTX, others don't.
    """
    _test_init(hlist,"Multiple TX")    
    
    for adu in range(NUM_ADU_PER_TEST):
        i = 0
        for h in hlist:
            mtx = i % 2
            data_send(h, adu, 0, 0, 0, mtx)
            i += 1

        dispatch_events(0)
        
        sleep(MTX_TEST_PERIOD)

    sleep(10) # drain stream queues
    dispatch_events(500)    
    
    _test_done(hlist)

def test_all(hlist):
    """
    priority = stream_index

    cgroup = ( adu_num + stream_index ) mod 3 + 1

    relative loss threshold of 1% on 'd'

    stream 'c' has MTX
    """
    _test_init(hlist,"integration")    
    
    for adu in range(NUM_ADU_PER_TEST):
        # 'a'
        cgroup = (adu + 0) % 3 + 1
        data_send(hlist[0], adu, 0, cgroup, 0, 0)        

        # 'b'
        cgroup = (adu + 1) % 3 + 1
        data_send(hlist[1], adu, 1, cgroup, 0, 0)

        # 'c'
        cgroup = (adu + 2) % 3 + 1
        data_send(hlist[2], adu, 2, cgroup, 0, 1)
        
        # 'd'
        cgroup = (adu + 3) % 3 + 1
        data_send(hlist[3], adu, 3, cgroup, 1, 0)

        dispatch_events(0)

        sleep(MTX_TEST_PERIOD)

    sleep(10)
    dispatch_events(500)    
    
    _test_done(hlist)

    
#
# server
#
def recv_cb(stream_handle, adu_id, totlen, offset, data):
    if data != ADU_DATA:
        #print "Bad ADU: %s" % (repr(data))
        print "!",

def do_server():
    if (horde_api.session_accept(SESSION_NAME)):
        print "connected to client."
        hlist = [None]*4        
        n = horde_api.stream_accept_multiple(4, STREAM_NAMES, hlist)
        print "accepted:", n, "streams"
        if (n == 4):
            for h in hlist: horde_api.stream_set_recv_callback(h, recv_cb)
            n = 0
            while (horde_api.event_dispatch() >= 0):
                if n%50 == 0:
                    sys.stdout.write(".")
                    sys.stdout.flush()
                n += 1
            print
        else:
            print "critical: wrong number of streams."

        stream_close_multiple(STREAM_NAMES, hlist)        
        horde_api.session_close(SESSION_NAME)
        print "session closed."
            
if len(sys.argv) == 1:
    # server
    do_server()
elif len(sys.argv) == 2:
    # client
    do_client(sys.argv[1])
else:
    print "usage:", sys.argv[0], "<server address>"
