import horde_api
import sys

SESSION_NAME = "pingpong"
STREAM_NAME = "pingpong"

hstream = 0

def session_cb(x):
    if (x == horde_api.EVENT_SESSION_CLOSED):
        print "session closed (%d). exiting." % x
        horde_api.session_close(SESSION_NAME)
        sys.exit(0)
    else:
        print "unknown session event:", x

def event_cb(event, stream_handle, a, b, c, d):
    if (event == horde_api.EVENT_ADU_FRAG_ACK):
        print "\tACK: stream=%d, adu=%d, offset=%d, fraglen=%d, rtt=%d" % (stream_handle, a, b, c, d)
        print "rtt:", d
    elif (event == horde_api.EVENT_ADU_FRAG_LOSS):
        print "\tLOSS: stream=%d, adu=%d, offset=%d, fraglen=%d" % (stream_handle, a, b, c)
        send_pingpong_packet(a);
    elif (event == horde_api.EVENT_STREAM_FLUSHED):
        print "\tFLUSH: stream flushed:", stream_handle
    elif (event == horde_api.EVENT_STREAM_CLOSED):
        print "\tCLOSED: stream closed:", stream_handle
    else:        
        print "\tEVENT: Unknown stream event [%d] on stream %d, a=%d, b=%d, c=%d, d=%d" % (event, stream_handle, a, b, c, d)

def send_pingpong_packet(seq):
    global hstream
    print "\tPONG: %d" % seq  
    horde_api.stream_send(hstream, seq, str(seq))

def recv_cb(stream_handle, adu_id, totlen, offset, data):
    global hstream
    
    print "PING [data=%s, adu=%d, totlen=%d, fraglen=%d, offset=%d]" % (repr(data), adu_id, totlen, len(data), offset)
    seq = int(data)    

    seq += 1
    send_pingpong_packet(seq);

def rate_to_kbps(rate):
    kbps = (rate * 8.0) / 1024;
    return "%.1f kbps" % kbps

def stat_cb(total_send, total_recv, total_allowed_send, nchannels):
    print "\tstatistics: send=%d (%s), recv=%d (%s), allowed=%d (%s), nchannels=%d"\
          % (total_send, rate_to_kbps(total_send), total_recv, rate_to_kbps(total_recv),
             total_allowed_send, rate_to_kbps(total_allowed_send), nchannels)

def horde_event_loop():
    while (horde_api.event_dispatch() > 0):
        pass

def do_client(server):
    global hstream
    
    if (horde_api.session_connect(server, SESSION_NAME)):
        print "connected to server '%s'." % server
        h = horde_api.stream_connect(STREAM_NAME)
        if (h > 0):
            hstream = h
            print "stream '%s' connected, uid:%d" % (STREAM_NAME, h)
            
            # register callbacks
            horde_api.session_set_callback(session_cb)
            horde_api.stream_set_event_callback(h, event_cb)
            horde_api.stream_set_recv_callback(h, recv_cb)

            horde_api.statistics_set_callback(stat_cb)
            horde_api.statistics_enable(1)

            # send the first ping
            send_pingpong_packet(0)
            
            # enter event loop
            horde_event_loop()            
    
def do_server():
    global hstream
    
    if (horde_api.session_accept(SESSION_NAME)):
        print "client connected."
        h = horde_api.stream_accept(STREAM_NAME)
        if (h > 0):
            hstream = h
            print "stream '%s' connected, uid:%d" % (STREAM_NAME, h)
            
            # register callbacks
            horde_api.session_set_callback(session_cb)
            horde_api.stream_set_event_callback(h, event_cb)
            horde_api.stream_set_recv_callback(h, recv_cb)

            horde_api.statistics_set_callback(stat_cb)
            horde_api.statistics_enable(1)

            # client will send a ping...

            # enter event loop
            horde_event_loop()
            
if len(sys.argv) == 1:
    # server
    do_server()
elif len(sys.argv) == 2:
    # client
    do_client(sys.argv[1])
else:
    print "usage:", sys.argv[0], "<server address>"
