import horde_api
import sys
from horde_ex import ReliableOrderedStream
from time import sleep

SESSION_NAME = "reliable"
STREAM_NAME = "reliable"

hstream = 0

SEND_SLEEP_DELAY = 1.0
ADU_SIZE = 16

def session_cb(x):
    if (x == horde_api.EVENT_SESSION_CLOSED):
        print "session closed (%d). exiting." % x
        horde_api.session_close(SESSION_NAME)
        sys.exit(0)
    else:
        print "unknown session event:", x

def rate_to_kbps(rate):
    kbps = (rate * 8.0) / 1024;
    return "%.1f kbps" % kbps

def stat_cb(total_send, total_recv, total_allowed_send, nchannels):
    print "\tstatistics: send=%d (%s), recv=%d (%s), allowed=%d (%s), nchannels=%d"\
          % (total_send, rate_to_kbps(total_send), total_recv, rate_to_kbps(total_recv),
             total_allowed_send, rate_to_kbps(total_allowed_send), nchannels)

def horde_event_loop():
    while (horde_api.event_dispatch() > 0):
        pass

def make_adu(adu_id, n):
    return str(range(adu_id, adu_id+n))

def close_cb(hstream):
    print "STREAM CLOSED:", hstream

def recv_cb(hstream, adu_id, data):
    print "RECV: stream=%d, adu=%d, data=\'%s\'" % (hstream, adu_id, data)
        
def do_client(server):
    global hstream
    
    if (horde_api.session_connect(server, SESSION_NAME)):
        print "connected to server '%s'." % server
        h = horde_api.stream_connect(STREAM_NAME)
        if (h > 0):
            hstream = h
            print "stream '%s' connected, uid:%d" % (STREAM_NAME, h)
            
            # register general callbacks
            horde_api.session_set_callback(session_cb)            
            horde_api.statistics_set_callback(stat_cb)
            horde_api.statistics_enable(1)

            # reliable stream
            rstream = ReliableOrderedStream(hstream, recv_cb, close_cb, True)

            # enter the send loop
            adu_id = 0
            while True:
                # send some data
                rstream.send(adu_id, make_adu(adu_id, ADU_SIZE), 0)
                adu_id += 1

                # dispatch some events
                while (horde_api.event_dispatch(0) > 0): pass

                # wait a bit
                sleep(SEND_SLEEP_DELAY)
                
    
def do_server():
    global hstream
    
    if (horde_api.session_accept(SESSION_NAME)):
        print "client connected."
        h = horde_api.stream_accept(STREAM_NAME)
        if (h > 0):
            hstream = h
            print "stream '%s' connected, uid:%d" % (STREAM_NAME, h)
            
            # register callbacks
            horde_api.session_set_callback(session_cb)
            horde_api.statistics_set_callback(stat_cb)
            horde_api.statistics_enable(1)

            # reliable stream
            rstream = ReliableOrderedStream(hstream, recv_cb, close_cb, False)            

            # enter event loop and receive data
            horde_event_loop()
            
if len(sys.argv) == 1:
    # server
    do_server()
elif len(sys.argv) == 2:
    # client
    do_client(sys.argv[1])
else:
    print "usage:", sys.argv[0], "<server address>"
