#ifndef __TRIBE_FOLLOWER_POOLMAN_H__
#define __TRIBE_FOLLOWER_POOLMAN_H__

#include <nmstl/ioevent>
#include <nmstl/net>

using namespace std;
using namespace nmstl;

#include "string_hashmap.h"
#include "raw_io.h"

extern io_event_loop io_loop;

#define POOLMAN_SLEEP_INTERVAL 5000

typedef void (*iface_event_handler)(string /*iface*/, inet_address /*address*/, bool up_event);

//
// interface pool manager
//
class interface_pool_manager {
 private:
  // the scheduling period
  uint32_t period_msecs;

  // is this poolman running?
  bool running;

  // list of managed interfaces
  vector<string> ifaces;

  // interface information snapshot
  STRING_HASHMAP(inet_address) iface_addresses;

  // I/O handlers
  STRING_HASHMAP(ptr<rawio_handler_group>) iface_iohandlers;
  hash_map< int, ptr<rawio_handler_group> > addr_to_iohandler_map;

  // event handlers
  iface_event_handler if_event_handler;

  // internal functions to handle pool management...
  callback_timer *wakeup_timer;


  // wakeup, examine state of interfaces, determine what needs to be done
  // and do it...
  void wakeup();

  // note that an interface came online, init the io handlers
  void iface_up (int idx, inet_address addr);

  // note that an interface went off, destroy the io handlers
  void iface_down(int idx);
  
 public:
  // constructor
  interface_pool_manager(vector<string>& managed_ifaces, uint32_t period, iface_event_handler handler=NULL) 
    : period_msecs(period), running(false), ifaces(managed_ifaces), iface_addresses(), 
    iface_iohandlers(), addr_to_iohandler_map(), if_event_handler(handler), wakeup_timer(NULL)
    { }

    bool is_running() { return running; }

    // starts pool manager, unless is_running().
    void start();  

    // stops, and deactivates all channel managers and io handlers
    void stop();

    // sends a packet using the correct raw socket.
    // async send. non-blocking.
    void send_raw(dynbuf& packet);    
};

#endif
