#include <vector>
#include <sys/types.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "ifmanager.h"
#include "util.h"

#include <nmstl/debug>

using namespace std;
using namespace nmstl;

void(*exit_handler)() = NULL;
vector<int> child_pid;

string get_ifman_cmdline(string& iface) {
  char *cwd = get_current_dir_name();
  string cmd = cwd;

  cmd += "/ifmanager";

  cmd += " --iface=";
  cmd += iface;
  cmd += " --max-retry=";
  cmd += to_string(if_max_retry);
  cmd += " --retry-reset-period=";
  cmd += to_string(if_retry_reset_period);
  cmd += " --retry-sleep=";
  cmd += to_string(if_retry_sleep);
  if (if_reset_all) cmd += " --reset-all";

  return cmd;
}

static void kill_children() {
  // destroy any child processes...
  char *cwd = get_current_dir_name();
  int pid = (int)getpid();

  char cmdbuffer[512];
  snprintf(cmdbuffer, 512, "%s/scripts/kill_all_children %d -recursive", cwd, pid);

  string com = cmdbuffer;
  spawn(com);
  waitpid(-1, NULL, 0);
}

void ifmanagers_start_all(vector<string>& managed_ifaces, void(*handler)()) {
  // set up event handler
  exit_handler = handler;

  // determine commandline
  
  // try to create and start all interface managers
  child_pid.clear();
  for (vector<string>::iterator i = managed_ifaces.begin(); i != managed_ifaces.end(); i++) {
    string com = get_ifman_cmdline(*i);    
    int pid = spawn(com);

    if (pid > 0) {
      child_pid.push_back(pid);
    } else {
      // failure...
      ifmanagers_kill_all(); // kill all the active guys...
      exit_handler(); // notify exit handler
      return;
    }
  }
}

void ifmanagers_kill_all() {
  kill_children();
  child_pid.clear();
}

void ifmanagers_poll() {
  int pid = waitpid(-1, NULL, WNOHANG);
  if (pid > 0) {
    if (exit_handler) { // somebody else cares about this event
      // a child exited, but it could be one of the children we explicitly killed...
      bool expect_alive=false;
      for (int i = 0; i < child_pid.size(); i++) 
	if (child_pid[i] == pid) {
	  expect_alive = true;
	  break;
	}
      
      if (expect_alive) exit_handler();
    }
  }
}
