#include <linux/netfilter.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/socket.h>	
#include <netinet/in.h>	
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include <stdio.h>
#include <stdlib.h>

extern "C" {
#include <libipq.h>
}

#include "util.h"

//#include <nmstl/debug>
//using namespace nmstl;

#define BUFSIZE 2048
unsigned char pkt_buf[BUFSIZE];
ipq_handle *ipq_h;
void (*packet_recv_handler)(unsigned char* /*data*/, size_t /*len*/)=NULL;
bool stealer_running=false;
pthread_t stealer;
vector<string> stolen_protos;

void *packet_stealer_run(void* unused) {
  int status, type, errcount=0;

  DEBUG << "packet stealer started.";

  while (stealer_running) {
    status = ipq_read(ipq_h, pkt_buf, BUFSIZE, 500000); // 500 msec timeout
    if (status < 0) {
      ipq_destroy_handle(ipq_h);
      ipq_perror("packet_stealer_run");
      break;
    } else if (status > 0) {
      type = ipq_message_type(pkt_buf);
      if (type == NLMSG_ERROR) {
	int err = ipq_get_msgerr(pkt_buf);
	if ((err != EAGAIN) && (err != EBUSY)) {
	  WARN <<  "Received unexpected NETFILTER error message " << err;
	  errcount++;
	  if (errcount > 32) {
	    CRIT << "Too many consecutive errors. Exiting packet stealer.";
	    break;
	  }
	}
      } else if (type == IPQM_PACKET) {
	errcount = 0;
	ipq_packet_msg_t *m = ipq_get_packet(pkt_buf);
	if (m->data_len) {
	  if (packet_recv_handler) packet_recv_handler(m->payload, m->data_len);  // pass to handler
	  status = ipq_set_verdict(ipq_h, m->packet_id, NF_DROP, 0, NULL);  // yoink...
	}
      }    
    }
  }

  ipq_destroy_handle(ipq_h);
  DEBUG << "packet stealer stopped.";
}

bool packet_stealer_start(void (*packet_recv_cb)(unsigned char* /*data*/, size_t /*len*/), const char* proto_list) {
  // set up iptables queueing rules etc
  int ret = system("modprobe ip_queue");
  INFO << "executing: modprobe ip_queue -> " << ret;
  if (ret) return false;

  stolen_protos = parse_list(proto_list);
  if (stolen_protos.size() == 0) return false;

  // initialize netfilter
  int status=0;
  ipq_h = ipq_create_handle(0, PF_INET);  
  if (ipq_h) status =  ipq_set_mode(ipq_h, IPQ_COPY_PACKET, BUFSIZE);
  if (!ipq_h || status<0) {
    ipq_destroy_handle(ipq_h);
    ipq_perror("start_packet_stealer"); 
    return false;
  } 

  // start tcp stealer thread
  stealer_running=true;
  ret = pthread_create(&stealer, NULL, packet_stealer_run, NULL);
  if (ret) {
    CRIT << "could not create packet stealer thread." << strerror(errno) << " [" << errno << "]";
    return false; 
  }

  packet_recv_handler = packet_recv_cb;

  return true;
} 

void packet_stealer_stop() {
  stealer_running=false;
}

void iptables_setup(string iface, bool add) {
  const char *addordel = (add ? "A" : "D");
  char commandbuf[256];
  int err = 0;

  if (stolen_protos[0] != "*") {
    for (int j = 0; j < stolen_protos.size(); j++) {
      snprintf(commandbuf, 256, "/sbin/iptables -%s INPUT -p %s -j QUEUE -i %s", addordel, stolen_protos[j].c_str(), iface.c_str());
      err |= system(commandbuf);
      INFO << "executing: " << commandbuf << "  -> " << err;
    }
  } else {
    snprintf(commandbuf, 256, "/sbin/iptables -%s INPUT -j QUEUE -i %s", addordel, iface.c_str());
    INFO << "executing: " << commandbuf << "  -> " << err;
    err |= system(commandbuf);      
  }
  if (err) CRIT << "there was an IPTABLES related problem. you may not receive any packets.";
}

void packet_stealer_ifup(string iface) {
  iptables_setup(iface, true);
}

void packet_stealer_ifdown(string iface) {
  iptables_setup(iface, false);
}

//
// random stuff (todo: move somewher else)
//
typedef iphdr   header_ip;
typedef icmphdr header_icmp;
typedef tcphdr  header_tcp;
typedef udphdr  header_udp;

string address_int32_to_string(u_int32_t addr);

int swab_short(short x) {
  int ret = 0;
  swab(&x, &ret, sizeof(short));
  return ret;
}

int swab_int(int x) {
  int ret = 0;
  swab(&x, &ret, sizeof(int));
  return ret;
}

string packet_header_to_string(char *pdata, size_t data_length, int offset) {
  //DEBUG << "enter: packet_header_to_string";

  pdata += offset;

  const int WORK_BUF_LEN = 1024;
  char workbuf[WORK_BUF_LEN];

  string retval("IP ");      

  // IP header     
  header_ip* ip_h = (header_ip*)pdata;

  snprintf(workbuf, WORK_BUF_LEN,
	   "[id=%d, src=%s, dest=%s, protocol=%d, tot_len=%d, ttl=%d]", 
	   (int)(swab_short(ip_h->id)), 
	   address_int32_to_string(ip_h->saddr).c_str(),
	   address_int32_to_string(ip_h->daddr).c_str(),
	   (int)(ip_h->protocol),
	   (int)(swab_short(ip_h->tot_len)),
	   (int)(ip_h->ttl));
  retval.append(workbuf);
  //DEBUG << "checkpoint: " << retval;
      
  if (ip_h->protocol == IPPROTO_TCP) {
    header_tcp* tcp_h = (header_tcp*)(pdata + sizeof(header_ip));

    snprintf(workbuf, WORK_BUF_LEN, 
	     ", TCP [seq=%d, port-src=%d, port-dst=%d",
	     (int)(swab_int(tcp_h->seq)),
	     (int)(swab_short(tcp_h->source)),
	     (int)(swab_short(tcp_h->dest)));

    retval.append(workbuf);

    if (tcp_h->urg) retval.append(", URG");
    if (tcp_h->ack) retval.append(", ACK");
    if (tcp_h->psh) retval.append(", PSH");
    if (tcp_h->rst) retval.append(", RST");
    if (tcp_h->syn) retval.append(", SYN");
    if (tcp_h->fin) retval.append(", FIN");
	
    retval.append("]");
	
  } else if (ip_h->protocol == IPPROTO_UDP) {
    header_udp* udp_h = (header_udp*)(pdata + sizeof(header_ip));

    snprintf(workbuf, WORK_BUF_LEN, 
	     ", UDP [port-src=%d, port-dst=%d]",
	     (int)(swab_short(udp_h->source)),
	     (int)(swab_short(udp_h->dest)));
    retval.append(workbuf);

  } else if (ip_h->protocol == IPPROTO_ICMP) {
    header_icmp* icmp_h = (header_icmp*)(pdata + sizeof(header_ip));
    snprintf(workbuf, WORK_BUF_LEN, 
	     ", ICMP [echo_seq=%d]", (int)(icmp_h->un.echo.sequence));
    retval.append(workbuf);
  } else { // unknown protocol
    //retval.append("");
  }
      
  snprintf(workbuf, WORK_BUF_LEN,
	   ", len=%d", 
	   (int)data_length);
  retval.append(workbuf); 

  //DEBUG << "exit: packet_header_to_string : " << retval;
  return retval;
}
