#include <sys/types.h>
#include <sys/socket.h>
#include <sys/socket.h>		
#include <netinet/ip.h>
#include <list>
#include <ext/hash_map>
#include <nmstl/ioevent>
#include <string.h>

#include <netinet/in.h>	
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include "tun_manager.h"

using namespace __gnu_cxx; // for hashmap

//
// utility functions etc
//
#define MAX_PACKET_LEN 2048
#define TUN_DEVICE_MTU 1400

int tun_device_activate(char* dev, const char* ip_address, int mtu=TUN_DEVICE_MTU);
int tun_device_deactivate(const char* dev, int fd);

string hex_dump(dynbuf& buf, int numshorts) {
  string hex="";
  char strbuf[16];
  //unsigned short *pbuf = (unsigned short *) buf.data();  

  unsigned short *pbuf = new unsigned short[numshorts];
  swab(buf.offset_data(), (char*)pbuf, numshorts*sizeof(short));

  for (int i = 0; i < numshorts; i++) {
    snprintf(strbuf, 16, "%04x ", pbuf[i]);

    hex.append(strbuf);  
  }

  return hex;
}

u_int32_t address_nmstl_to_int32(const inet_address& addr) {
  // ugh. inefficient hack: nmstl -> string -> int32
  string addrstr = addr.get_host();
  u_int32_t addrint = inet_addr(addrstr.c_str());
  return addrint;
}

string address_int32_to_string(u_int32_t addr) {      
  struct in_addr saddr;
  saddr.s_addr = addr;
  //swab(&addr, &(saddr.s_addr), 4);
  
  char buf[INET_ADDRSTRLEN+1];
  inet_ntop(AF_INET, &(saddr), buf, INET_ADDRSTRLEN);
  return string(buf);
}

typedef iphdr   header_ip;
typedef icmphdr header_icmp;
typedef tcphdr  header_tcp;
typedef udphdr  header_udp;

int swab_short(short x) {
  int ret = 0;
  swab(&x, &ret, sizeof(short));
  return ret;
}

int swab_int(int x) {
  int ret = 0;
  swab(&x, &ret, sizeof(int));
  return ret;
}

string packet_header_to_string(char *pdata, size_t data_length) {
  //DEBUG << "enter: packet_header_to_string";

  pdata += 4;

  const int WORK_BUF_LEN = 1024;
  char workbuf[WORK_BUF_LEN];

  string retval("IP ");      

  // IP header     
  header_ip* ip_h = (header_ip*)pdata;

  snprintf(workbuf, WORK_BUF_LEN,
	   "[id=%d, src=%s, dest=%s, protocol=%d, tot_len=%d, ttl=%d]", 
	   (int)(swab_short(ip_h->id)), 
	   address_int32_to_string(ip_h->saddr).c_str(),
	   address_int32_to_string(ip_h->daddr).c_str(),
	   (int)(ip_h->protocol),
	   (int)(swab_short(ip_h->tot_len)),
	   (int)(ip_h->ttl));
  retval.append(workbuf);
  //DEBUG << "checkpoint: " << retval;
      
  if (ip_h->protocol == IPPROTO_TCP) {
    header_tcp* tcp_h = (header_tcp*)(pdata + sizeof(header_ip));

    snprintf(workbuf, WORK_BUF_LEN, 
	     ", TCP [seq=%d, port-src=%d, port-dst=%d",
	     (int)(swab_int(tcp_h->seq)),
	     (int)(swab_short(tcp_h->source)),
	     (int)(swab_short(tcp_h->dest)));

    retval.append(workbuf);

    if (tcp_h->urg) retval.append(", URG");
    if (tcp_h->ack) retval.append(", ACK");
    if (tcp_h->psh) retval.append(", PSH");
    if (tcp_h->rst) retval.append(", RST");
    if (tcp_h->syn) retval.append(", SYN");
    if (tcp_h->fin) retval.append(", FIN");
	
    retval.append("]");
	
  } else if (ip_h->protocol == IPPROTO_UDP) {
    header_udp* udp_h = (header_udp*)(pdata + sizeof(header_ip));

    snprintf(workbuf, WORK_BUF_LEN, 
	     ", UDP [port-src=%d, port-dst=%d]",
	     (int)(swab_short(udp_h->source)),
	     (int)(swab_short(udp_h->dest)));
    retval.append(workbuf);

  } else if (ip_h->protocol == IPPROTO_ICMP) {
    header_icmp* icmp_h = (header_icmp*)(pdata + sizeof(header_ip));
    snprintf(workbuf, WORK_BUF_LEN, 
	     ", ICMP [echo_seq=%d]", (int)(icmp_h->un.echo.sequence));
    retval.append(workbuf);
  } else { // unknown protocol
    //retval.append("");
  }
      
  snprintf(workbuf, WORK_BUF_LEN,
	   ", len=%d", 
	   (int)data_length);
  retval.append(workbuf); 

  //DEBUG << "exit: packet_header_to_string : " << retval;
  return retval;
}

extern io_event_loop io_loop;

//
// manager for a file descriptor on a tun device
//
class tun_device_iohandler : io_handler {
private:
  list<dynbuf> write_queue;
  inet_address dev_address;
  inet_address fwd_address;
  int devfd;
  string devname;
  int tunidx;
  tundev_recv_cb recv_handler;
public:
  // Create a tun device handler from a file descriptor to a tun device interface
  // and an address for that interface.
  tun_device_iohandler(int fd, string name, int tunidx,
		       const inet_address& dev_address, 
		       const inet_address& fwd_address,
		       tundev_recv_cb hrecv) :
    io_handler(io_loop, iohandle(fd)), write_queue(), devfd(fd), devname(name), tunidx(tunidx),
    dev_address(dev_address), fwd_address(fwd_address), recv_handler(hrecv) { 
    want_read(true);
    want_write(false);
    get_ioh().set_blocking(false);

    DEBUG << "tun_device_iohandler created: " << name;
  }

  void write_packet(dynbuf& buf) {
    write_queue.push_back(buf);
    want_write(true);
  }

  string get_devname() { return devname; }
  int get_devfd() { return devfd; }
  int get_tunidx() { return tunidx; }

  ~tun_device_iohandler() {
    close();
    DEBUG << "tun_device_iohandler destroyed: " << devname;
  }

protected:
  void ravail() {
    dynbuf buf(MAX_PACKET_LEN);
    ssize_t n = ::read(devfd, buf.data(), buf.max_length());

    if (n > 0) {
      buf.set_length(n);
      recv_handler(buf, fwd_address);

      DEBUG(3) << "tun_device_iohandler::read from " << devname 
	       << ": " << packet_header_to_string(buf.data(), n);
    } else {
      WARN << "tun_device_iohandler::read error on "  << devname;
    }
  }

  void wavail() {
    if (write_queue.empty()) {
      want_write(false);    
    } else {
      dynbuf buf = write_queue.front();

      int len = buf.length();
      ssize_t n = ::write(devfd, buf.data(), len);

      if (n < len)
	WARN << "tun_device_iohandler::write on " << devname
	     << " some bytes dropped: written=" << n << ", total=" << len;
      else
	DEBUG(3) << "tun_device_iohandler::write, wrote to " << devname 
		 << ": " << packet_header_to_string(buf.data(), n);;
      
      write_queue.pop_front();
    }
  }
};

//
// helper class to track which TUN devices are active
//
class activation_bitmap {
private:
  static const int BITMAP_SIZE = 256;
  bool bitmap[256];
public:
  activation_bitmap() { for (int i = 0; i < BITMAP_SIZE; i++) bitmap[i] = false;  }
  
  void set_active(int idx) { if ((idx < BITMAP_SIZE) && (idx >= 0)) bitmap[idx] = true; }
  void set_inactive(int idx) { if ((idx < BITMAP_SIZE) && (idx >= 0)) bitmap[idx] = false; }

  bool is_active(int idx) { 
    if ((idx < BITMAP_SIZE) && (idx >= 0)) 
      return bitmap[idx]; 
    else
      return false;
  }

  /// returns lowest index for which isActive would return false, or -1 if 
  /// no such index can be found.
  int lowest_inactive() {    
    for (int i = 0; i < BITMAP_SIZE; i++) if (!bitmap[i]) return i;
    return -1;
  }
};

activation_bitmap tun_dev_tracker;

//
// exported functions
//

// map of: IPv4-addr in net order -> iohandler
hash_map<u_int32_t, tun_device_iohandler* > addr_to_dev_map;

void destroy_dev_handler(tun_device_iohandler* devh) {
  string key = devh->get_devname();
  int tunidx = devh->get_tunidx();

  // deactivate the device
  tun_device_deactivate(devh->get_devname().c_str(), devh->get_devfd());

  // remove from maps
  for (hash_map< u_int32_t, tun_device_iohandler* >::iterator iter = addr_to_dev_map.begin();
       iter != addr_to_dev_map.end();
       iter++) {
    if (iter->second == devh) {
      addr_to_dev_map.erase(iter);
      break;
    }
  }

  tun_dev_tracker.set_inactive(tunidx);

  // delete
  delete devh;

  INFO << "TUN device destroyed: " << key;
}

tun_device_iohandler* create_dev_handler(string name, string devaddr, inet_address tunnel_dest_addr, tundev_recv_cb hrecv) {
  inet_address iface_addr(devaddr);
  tun_device_iohandler* ph = NULL;

  // create a tun device
  char devname[32], ipaddr[32];
  int tunidx = tun_dev_tracker.lowest_inactive();
  snprintf(devname, 32, "tun%d\0", tunidx);
  strcpy(ipaddr, devaddr.c_str());

  int devfd = tun_device_activate(devname, ipaddr);

  if (devfd >= 0) {
    // create a new io handler
    ph = new tun_device_iohandler(devfd, string(devname), tunidx, iface_addr, tunnel_dest_addr, hrecv);

    // set up maps correctly...
    addr_to_dev_map[address_nmstl_to_int32(iface_addr)] = ph;
    tun_dev_tracker.set_active(tunidx);
    
    // log success
    INFO << "TUN device active: remoteid=" << name << ", local=" << devname << ", fd=" << ph->get_devfd();
  } else {
    ERROR << "could not create TUN device for: " << name << ", fd=" << devfd << " [attempted devname=" << devname << "]"; 
  }

  return ph;
}

void route_to_tun_device(dynbuf& packet) {
  // Data contains an IP address, peek IP header and figure out which
  // TUN device this was meant for...
  iphdr* hdr = (iphdr*)(packet.offset_data() + 4);
  
  hash_map<u_int32_t, tun_device_iohandler*>::iterator pphandler = addr_to_dev_map.find(hdr->daddr);
  
  if (pphandler != addr_to_dev_map.end()) {
    // Tell that TUN-device's io handler to handle this packet...
    (pphandler->second)->write_packet(packet);
  } else {
    WARN << "route_to_tun_device: dropping packet to " << address_int32_to_string(hdr->daddr);
    DEBUG << hex_dump(packet, packet.length() / 2);
  }
}

